/* nylon - (socks{4,5}) proxy and more
 * http://mesh.eecs.umich.edu/projects/nylon
 * marius aamodt eriksen (marius@umich.edu) */

/* XXX i should probably break this down into more manageable source
 * files */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/signal.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#if defined(__sun__)
#include <sys/sockio.h>
#endif /* __sun__ */

#include <netdb.h>
#include <netinet/in.h>
#include <net/if.h>
#include <errno.h>

#include <stdio.h>
#include <unistd.h>

#include "mnet.h"
#include "cfg.h"
#include "util.h"
#include "config.h"

#ifdef WIN32
#include "getopt.h"
#endif /* WIN32 */


#define DEFAULT_MALLOC 16   /* userids usually aren't that large */
#define IFMAX IFNAMSIZ
#define S4_MAX_USERID 256

/* socks4 request/reply header */
typedef struct {
	u_char vn;
	u_char cd;
	u_int16_t dport;
	u_int32_t dip;
} socks4_hdr;
#define SOCKS4_HDR_LEN ((2*1)+2+4)

/* socks5 request/reply header */
typedef struct {
	u_char  vn;
	u_char  cd;
	u_char  rsv;
	u_char  atyp;
	u_int32_t daddr;
	u_int16_t dport;
} socks5_req;
#define SOCKS5_REQ_LEN ((4*1)+4+2)

/* version reply */
typedef struct {
	u_char ver;
	u_char res;
} socks5_vrepl;
#define SOCKS5_VREPL_LEN (2*1)

typedef struct {
	char *if_bind;
	char *if_conn;
	struct ip_chain *ip_allow;
	struct ip_chain *ip_deny;
	char *pidfile;
	short port;
	int numconnects;
	int log;
} nylon_conf;

char *cvsid = "$Id: nylon.c,v 1.17 2001/10/07 07:17:55 beriksen Exp $";

int verbose, active_connections;

/* XXX */
char *pidfile;

#define VERSION "0.2g"

void usage __P((char**));
void wait_for_connections __P((nylon_conf*));
void negotiate_connection __P((nylon_conf*, int, int));
void negotiate4 __P((nylon_conf*, int, int));
void negotiate5 __P((nylon_conf*, int, int));
void establish_connection __P((int, int, int));
void sigchld __P((int));
void sigterm __P((int));

/*
 * main() - read in command line options and act upon them, then
 * fork-and-exit for ``daemon like'' behavior
 */
int
main (int argc, char **argv)
{
	nylon_conf conf;
	int opt, foreground, bgproc;
	char *_ip_allow, *_ip_deny;
	char *_inf, *_cinf, /* *_laddr,*/ apid[6]; 
	FILE *pidfd;

	/* defaults */

	active_connections = conf.log = verbose = foreground = 0;
	conf.numconnects = 10;
	conf.port = 1080;
	conf.if_bind = conf.if_conn = conf.pidfile = NULL;
	pidfile = conf.pidfile;

	/* take care of configuration */

	conf_init();
	CONF_SAVE(conf.if_bind, conf_get_str("Server", "Binding-Interface"));
	CONF_SAVE(conf.if_conn, conf_get_str("Server", "Connecting-Interface"));
	CONF_SAVE(_ip_allow, conf_get_str("Server", "Allow-IP"));
	CONF_SAVE(_ip_deny, conf_get_str("Server", "Deny-IP"));
/*	CONF_SAVE(cinf, conf_get_str("OnlyAccept", "Listenfor-Address"));*/
	CONF_SAVE(conf.pidfile, conf_get_str("General", "PIDFile"));	
	conf.port = conf_get_num("Server", "Port", 1080);
	conf.numconnects = conf_get_num("General", "No-Simultaneous-Conn", 10);
	conf.log = conf_get_num("General", "Log", 0);
	verbose = conf_get_num("General", "Verbose", 0);

	while ((opt = getopt(argc, argv, "hvVfp:n:i:I:")) != -1) 
 		switch (opt) {
		case 'h':
			usage(argv);
		case 'v':
			verbose++;
			break;
		case 'V':
			printf("nylon version " VERSION ":\n%s\n", cvsid);
			exit(0);
		case 'p':
			conf.port = atoi(optarg);
			break;
		case 'n':
			conf.numconnects = atoi(optarg);
			break;
		case 'i':
			_inf = (char*)malloc(strlen(optarg)+1);
			strlcpy(_inf, optarg, strlen(optarg)+1);
			conf.if_bind = _inf;
			printv(1, "binding with %s\n", conf.if_bind);
			break;
		case 'I':
			_cinf = (char*)malloc(strlen(optarg)+1);
			strlcpy(_cinf, optarg, strlen(optarg)+1);
			conf.if_conn = _cinf;
			printv(1, "binding with %s\n", conf.if_conn); 
			break;
		case 'f':
			foreground = 1;
			break;
		default:
			break;
		}

	/* set up allow/deny chains */

	conf.ip_allow = (struct ip_chain*)(_ip_allow ? make_ipchain(expanda(_ip_allow))
									   : NULL);
	conf.ip_deny  = (struct ip_chain*)(_ip_deny ? make_ipchain(expanda(_ip_deny))
									   : NULL);
	printv(1, "ip allow chain %sinitialized\n", conf.ip_allow ? "":"not ");
	printv(1, "ip deny chain %sinitialized\n", conf.ip_deny ? "":"not ");	

	bgproc = 0;

	if (!foreground)
		bgproc = fork();

	if (bgproc < -1) 
		errx(1, "fork() failed");
	if (bgproc != 0) {
		printv(0, "nylon launched with pid %i\n", bgproc);
		if (conf.pidfile) {
			if (!(pidfd = fopen(conf.pidfile, "w")))
				errx(1, "could not write to pid file");
			snprintf(apid, 6, "%i", bgproc);
			fputs(apid, pidfd);
			fclose(pidfd);
		}
		exit(0);
	}
	/* child */
	wait_for_connections(&conf);
	/* NOTREACHED */
	exit(0);
}

void
wait_for_connections(nylon_conf *conf)
{
	struct sockaddr_in serv_addr, cli_addr;
	int sfd, nsfd, nproc, session_no = 0;
	socklen_t cli_len = sizeof(struct sockaddr);
	char ifip[128];
	sigset_t ss_chld;

	sigemptyset(&ss_chld);
	sigaddset(&ss_chld, SIGCHLD);

	signal(SIGCHLD, sigchld);    /* get 'em zombies away! */
	signal(SIGTERM, sigterm);    /* graceful exit         */

	bzero(&serv_addr, sizeof(serv_addr));
	bzero(&cli_addr , sizeof(cli_addr));

	if ((sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		errx(1, "socket() failed");

	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port   = htons(conf->port);

	if (conf->if_bind) {
#ifndef WIN32
		serv_addr.sin_addr = get_ip_from_if(conf->if_bind);
	} else
		serv_addr.sin_addr.s_addr = INADDR_ANY;
#else
	  struct hostent *infhost;
	  infhost = gethostbyname(conf->if_bind);
	  serv_addr.sin_addr = *(struct in_addr*)infhost->h_addr;
	} else
		serv_addr.sin_addr.s_addr = INADDR_ANY;
#endif
		
	if (bind(sfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
		errx(1, "bind() failed");
	if (listen(sfd, conf->numconnects) == -1)
		errx(1, "listen() failed");
	printv(1, "listening...\n");
	for (;;) {
		if ((nsfd = accept(sfd, (struct sockaddr*) &cli_addr, &cli_len)) == -1 )
			errx(1, "accept() failed");
		printv(1, "accept()ed\n");
		sigprocmask(SIG_BLOCK, &ss_chld, NULL);
		if (active_connections > conf->numconnects-1)
			continue;
		sigprocmask(SIG_UNBLOCK, &ss_chld, NULL);
		session_no++;
		nproc = fork();
		if (nproc == -1)
			warnx(1, "fork() failed");
		if (nproc == 0) {
			printv("connect! starting negotiations\n");
			negotiate_connection(conf, nsfd, session_no);
			exit(0);
		}
		close(nsfd);
		sigprocmask(SIG_BLOCK, &ss_chld, NULL);		
		active_connections++;
		sigprocmask(SIG_UNBLOCK, &ss_chld, NULL);		
	}
}

/* multiplex depending on version number */
void
negotiate_connection(nylon_conf *conf, int sfd, int session_no)
{
	unsigned char vn;

	if (recvn(sfd, &vn, 1, 0) == -1)
		errx(1, "error receiving version number\n");

	if (vn == 4)               /* SOCKS4 */
		negotiate4(conf, sfd, session_no);
	else if (vn == 5)          /* SOCKS5 */
		negotiate5(conf, sfd, session_no);
	else
		errx(1, "No support for requested protocol version\n");
}

void
negotiate4(nylon_conf *conf, int sfd, int session_no)
{
	unsigned char vn, cd, *userid, *methods, method;
	unsigned short dport;
	unsigned long dip;
	unsigned i = 0, j = 1, nmethods;
	struct sockaddr_in target_addr, connect_addr;
	int target_sfd;
	socks4_hdr hdr4;

	printv(1, "entering SOCKS4 specific negotiation\n");

	userid = (unsigned char*)malloc(DEFAULT_MALLOC);

	hdr4.vn = 4;
	/* getting the seven first bytes after version, until USERID */
	if (recvn(sfd, &hdr4.cd, 7, 0) == -1)
		errx(1, "error receiving header\n");

	printv(2, "getting user id\n");

	while (recvn(sfd, userid+i, 1, 0) != -1)
		if ((userid[i]) == 0) {
			userid[i] = '\0';
			break;
		} else {
			if ((++i)%DEFAULT_MALLOC == 0)
				userid = (unsigned char*)realloc(userid, (++j)*DEFAULT_MALLOC);
			if ((j*DEFAULT_MALLOC) > S4_MAX_USERID)
				errx(1, "header too long, bailing out");
		}
	
	if (hdr4.cd != 1)
		errx(1, "SOCKS4: unsupported command!\n");

	if ((target_sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		errx(1, "could not create socket");

	bzero(&target_addr, sizeof(target_addr));
	target_addr.sin_family      = AF_INET;
	target_addr.sin_port        = hdr4.dport;   /* already in network order */
	target_addr.sin_addr.s_addr = hdr4.dip;     /* ditto                    */

	if (conf->if_conn) {
#ifdef WIN32
	  struct hostent *cinfhost;
	  cinfhost = gethostbyname(conf->if_conn);
	  connect_addr.sin_addr = *(struct in_addr*)cinfhost->h_addr;
#else /* WIN32 */
		connect_addr.sin_addr = get_ip_from_if(conf->if_conn);
#endif /* WIN32 */
		connect_addr.sin_port = 0;
		if (bind(target_sfd, (struct sockaddr*) &connect_addr,
				  sizeof(connect_addr)) < 0)
			errx(1, "bind failed");
	}
		
	printv("2, trying to connect to target host");

	if (connect(target_sfd,
				(struct sockaddr*)&target_addr, sizeof(target_addr)) < 0) 
		hdr4.cd = 91;
	else
		hdr4.cd = 90;

	hdr4.vn = 0; 

	if ( sendn(sfd, &hdr4, sizeof(hdr4), 0) == -1 )
		errx(1, "error replying to initiating host");

	if (cd == 91)
		errx(1, "error connecting to target host");
	
	printv(1, "connected!\n");

	establish_connection(sfd, target_sfd, session_no);
	
	/* NOTREACHED */
	return;
}


void
negotiate5(nylon_conf *conf, int sfd, int session_no)
{
	unsigned i = 0, j = 1, reqlen = SOCKS5_REQ_LEN;
	struct sockaddr_in target_addr, connect_addr;
	int target_sfd;
	socks5_req req5;
	socks5_vrepl rep5;
	unsigned char nmethods, *methods, method; 
	
	printv(1, "initiating SOCKS5 connection\n");

	method   = 0;
	req5.vn  = 5;
	req5.rsv = 0;

	/* start by retrieving number of methods, version number has
     * already been consumed by the calling procedure */

	if (recvn(sfd, &nmethods, 1, 0) == -1)
		errx(1, "error in receiving header");
	printv(2, "client advertises %i methods\n", nmethods);
	methods = (unsigned char*)malloc(nmethods);
	if (recvn(sfd, methods, nmethods, 0) == -1)
		errx(1, "error in receiving header");
	printv(2, "got all methods\n");

	/* we don't support any authentication methods yet, so simply
	 * ignore it and send reply with no authentication required. */
	
	free(methods);

	rep5.ver = 5;
	rep5.res = 0;
	
	printv(2, "sending reply: version: %i method: %i ...", req5.vn, method);
	if (sendn(sfd, &rep5, 2, 0) == -1)
		errx(1, "error in sending reply");
	printv(2, "sent\n");

	/* receive up to atyp */
	if (recvn(sfd, &req5, 4, 0) == -1)
		errx(1, "Error receiving SOCKS5 request");
	printv(2, "received the first 4 bytes of request\n");
	if (req5.vn != 5)
		errx(1, "Error in header");
	if (req5.cd != 1)
		errx(1, "Unsupported SOCKS5 request");
	printv(2, "client requested service %i\n", req5.cd);

	bzero(&target_addr, sizeof(target_addr));

	if (req5.atyp == 1) {   /* IPV4 Addr */
		printv(2, "ipv4 address requested\n");
		if (recvn(sfd, &req5.daddr, 4, 0) == -1)
			errx(1, "Error receiving request");
		target_addr.sin_family      = AF_INET;
		target_addr.sin_addr.s_addr = req5.daddr;  /* already in
                                                    * network order */
    } else if (req5.atyp == 3) {  /* fqdn */
		unsigned char len;
		char *hostname;
		struct hostent *hent;
		if (recvn(sfd, &len, 1, 0) == -1)
			errx(1, "error receiving request");
		hostname = (char*)malloc(len+1);
		if (recvn(sfd, hostname, len, 0) == -1)
			errx(1, "error receiving hostname");
		errx(1, "received hostname\n");
		hostname[len] = '\0';
		hent = gethostbyname(hostname);
		target_addr.sin_family = AF_INET;
		target_addr.sin_addr = *(struct in_addr*)hent->h_addr;
		free(hostname);
	} else 
		errx(1, "Unsupported address type"); /* XXX - die properly, sending
											  * error code back to client */

	if (recvn(sfd, &req5.dport, 2, 0) == -1)
		errx(1, "Error receving request");
	target_addr.sin_port = req5.dport;

	printv(1, "target host:\n");
	printv(1, "\tip address: %s\n", inet_ntoa(target_addr.sin_addr));
	printv(1, "\tport      : %i\n", ntohs(req5.dport));

	/* start connecting to target */

	if ((target_sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		errx(1, "could not create socket");

	printv(2, "making connection to target host...\n");

	if (conf->if_conn) {
#ifdef WIN32
	  struct hostent *cinfhost;
	  cinfhost = gethostbyname(conf->if_conf);
	  connect_addr.sin_addr = *(struct in_addr*)cinfhost->h_addr;
#else /* WIN32 */
	  connect_addr.sin_addr = get_ip_from_if(conf->if_conn);
#endif /* WIN32 */
	  connect_addr.sin_port = 0;
	  if (bind(target_sfd, (struct sockaddr*) &connect_addr,
				sizeof(connect_addr)) < 0)
		  errx(1, "bind failed");
	}
		
	if (connect(target_sfd,
				(struct sockaddr*)&target_addr, sizeof(target_addr)) < 0) {
		printv(1, "unsuccessful connect!\n");
		req5.cd = 1;
	} else {
		printv(1, "successful connect!\n");
		req5.cd = 0;
	}
	
	printv(2, "made connection to target host...\n");
	
	req5.atyp = 1; /* we are always sending an IPv4 addr back */ 
	
	if (sendn(sfd, &req5, 10, 0) == -1)
		errx(1, "error replying to initiating host");
	if (req5.cd == 1)
		errx(1, "error connecting to target host");
	
	establish_connection(sfd, target_sfd, session_no);
	
	/* NOTREACHED */
	return;
}

void
establish_connection(int client, int target, int session_no)
{
	fd_set rfds;
	int nfds, ret;
	unsigned char buf[1024];
	unsigned bcount;

	nfds = (client > target?client:target)+1;
	
	for (;;) {
		FD_ZERO(&rfds);
		FD_SET(client, &rfds); FD_SET(target, &rfds);
		ret = select(nfds, &rfds, NULL, NULL, NULL);
		bzero(buf, 1024);
		if (FD_ISSET(client, &rfds)) {
			if ((bcount = recv(client, buf, 1024, 0)) == -1)
				errx(1, "error in recv()");
			if (bcount == 0)
				if (verbose)
					errx(1, "client terminated connection");
				else
					exit(0);
			printv(3, "session %i: %i bytes client->target\n", session_no,
				   bcount);
			if (sendn(target, buf, bcount, 0) == -1)
				errx(1, "error in send()");
		} else if (FD_ISSET(target, &rfds)) {
			if ((bcount = recv(target, buf, 1024, 0)) == -1)
				errx(1, "error in recv()");
			if (bcount == 0)
				if (verbose)
					errx(1, "server terminated connection");
				else
					exit(0);
			printv(1, "session %i: %i bytes target->client\n", session_no,
				   bcount);
			if (sendn(client, buf, bcount, 0) == -1)
				errx(1, "error in send()");			
		} else if (ret < 0)
			errx(1, "select() failed");
	}
}

void
sigterm (int sign)
{
	unlink(pidfile);
	exit(0);
}

void
sigchld (int sign)
{
	pid_t pid;
	int   stat;

	pid = wait(&stat);
	active_connections--;
	printv(1, "decremented active count, now %i\n", active_connections);
	return;
}

void
usage (char **argv)
{
#ifndef WIN32
#define IFIP "if"
#else /* WIN32 */
#define IFIP "ip" 
#endif /* WIN32 */

	printf("nylon version " VERSION " usage:\n"
		   "%s [-h] [-v] [-V] [-f] [-p portno] [-i " IFIP "] [-I if] \n"
		   "   -h         help (this)\n"
		   "   -v         verbose\n"
		   "   -V         version\n"
		   "   -f         run as foreground process\n"
		   "   -i if      bind to specified " IFIP " (listening)\n"
		   "   -I if      bind to specified " IFIP " (connecting)\n"
		   "   -l ip      only accept connections from specified ip address\n"
		   "   -p portno  port number\n", *argv);
	exit(0);
}

/* copyright (c) 2001  
 * marius aamodt eriksen
 * all rights reserved
 * 
 * permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any purpose,
 * so long as the name of the author is not used in any advertising or 
 * publicity pertaining to the use or distribution of this software 
 * without specific, written prior authorization.  if the above copyright 
 * notice or any other identification of the author is included in any 
 * copy of any portion of this software, then the disclaimer below must 
 * also be included.
 * 
 * this software is provided as is, without representation from the
 * the author as to its fitness for any purpose, and without warranty by 
 * the author of any kind, either express or implied, including without 
 * limitation the implied warranties of merchantability and fitness for 
 * a particular purpose.  the author shall not be liable for any damages,
 * including special, indirect, incidental, or consequential damages,
 * with respect to any claim arising out or in connection with the use
 * of the software, even if it has been or is hereafter advised of the
 * possibility of such damages. */
